#ifndef UIDATA_H
#define UIDATA_H

#include <iostream>
#include <string>
#include <sstream>
#include <cmath>
#include <vector>
#include <complex>
#include <boost/shared_ptr.hpp>
#include <boost/regex.hpp>
#include "libharkio/libharkio.h"
#include "libharkio/textdata.h"
#include <QObject>
#include "common.h"
#include "Matrix.h"
#include "UIException.h"
#include "UIDocument.h"

//! 内部計算用データ型(最終的なファイル出力より、精度の高い型にしている)
//\{
typedef double HT_DATA_TYPE;
typedef std::complex<HT_DATA_TYPE> HT_COMPLEX_DATA_TYPE;
//\}

namespace HT {
class UIDocument;

template <typename T>
class UISrc {
public:
	T x;
	T y;
	T z;
	//! 新規にTFファイルを作成する時だけに使用されるデータ
	//\{
	std::string implement_file;	//!srcinfテキストに埋め込まれたfltファイル名(harkioを使いない場合)
	int cut_start;				//! srcinfファイルのfltファイル名の後ろの数(harkioを使いない場合)
	//\}
	T search_direction; // harkioを使わない旧形式のデータに使用される(定位のみ)
	//!デバック関数
	void print() {
		std::cout << "x=" << x << " y=" << y << "  z=" << z << " file=" << implement_file << " cut_start=" << cut_start << std::endl;
	}
	void convertFromHarkIO(harkio::UISrc<> &src) {
		x = src.x;
		y = src.y;
		z = src.z;
	}
};

//! micfile用データの保存クラス
template <typename T>
class UIMic{
public:
	//! cartesian_polar用マスクビット CARTESIAN:phi,theta有効   POLAR:z,y,x有効
	typedef enum {CARTESIAN=4+2+1, POLAR=16+8} MASK;
	int id;  //! micfile に入っているインデックス番号
	T x;
	T y;
	T z;
	T theta; // (harkioを使いない場合)
	T phi; //(harkioを使いない場合)
	int cartesian_polar; //! デカルト座標、極座標表示を判定するフラグ(harkioを使いない場合)
	UIMic() : cartesian_polar(CARTESIAN) {}
	//! デバック関数
	void print() {
		std::cout << "id=" << id << " x=" << x << " y=" << y << " z=" << z  \
		<< " theta=" << theta << " phi=" << phi << " flag=" << cartesian_polar << std::endl;
	}
	void convertFromHarkIO(harkio::UIMic<> &mic) {
		x = mic.x;
		y = mic.y;
		z = mic.z;
	}
};

template <typename T>
class UINoise{
public:
	T x;
	T y;
	T z;
	UINoise() {
		x=0; y=0; z=0;
	}
	void convertFromHarkIO(harkio::UINoise<> &noise) {
		x = noise.x;
		y = noise.y;
		y = noise.z;
	}
};

//! Dataの抽象基底クラス
class UIData : public QObject {
Q_OBJECT

protected:
//	harkio::HARK* harkio_data; //! 読み込み時と新規作成時のlibharkioで作ったデータコンテナ(旧フォーマットの時はNULL)
public:
	UIDocument* document; //! このクラスを管理包含している、ドキュメントクラス
	UIData(){}
	virtual ~UIData(){}

	//! このクラウスを管理しているdocumentクラスのポインタを取得
	HT::UIDocument* getDocument();
	void setDocument(UIDocument* document);


signals:
	void createdTSP(int current, int total, HT::UIData* data);
};


// 新旧、型の変換クラス
class Transform {
	std::vector<std::vector<double> > mics;
	std::vector<std::vector<double> > srcs;
	TFMatrix<HT_COMPLEX_DATA_TYPE>* old_matrix;
public:
	Transform() {
		old_matrix = NULL;
	}
	~Transform() {
		delete old_matrix;
	}
	void BasicData(HT::UIData& data, int* mic, int* src, int* fft, int* fs);
	TFMatrix<HT_COMPLEX_DATA_TYPE>& commonTFMatrix(HT::UIData& data);
	TFMatrix<HT_COMPLEX_DATA_TYPE>& commonIFFTTFMatrix(HT::UIData& arg_data);
	// M2PG::TF_LIST_TYPE(HARKIO伝達関数) -> TFMatrix(旧形式)
	TFMatrix<HT_COMPLEX_DATA_TYPE>& HARKIOTFToTFMatirx(harkio::M2PG::TF_LIST_TYPE* tf);
	TFMatrix<HT_COMPLEX_DATA_TYPE>& HARKIOTFToTFMatirx(harkio::GTF::TF_LIST_TYPE* tf);
	// vector<UISrc<T> >(旧形式マイク) -> std::vector<std::vector<T> > (GNUPlot散布図)
	std::vector<std::vector<double> >& OldMicToGNUPlotVector(std::vector<HT::UIMic<HT_DATA_TYPE> >* mic);
	// vector<UISrc<T> >(旧形式ソース) -> std::vector<std::vector<T> > (GNUPlot散布図)
	std::vector<std::vector<double> >& OldSrcToGNUPlotVector(std::vector<HT::UISrc<HT_DATA_TYPE> >* src);
	// M2PG::SRC_LIST_TYPE(HARKIOマイク) -> std::vector<std::vector<T> > (GNUPlot散布図)
	std::vector<std::vector<double> >& HARKIOMicToGNUPlotVector(harkio::FMT::MIC_LIST_TYPE* mic);
	// FMT::MIC_LIST_TYPE(HARKIOソース) -> std::vector<std::vector<T> > (GNUPlot散布図)
	std::vector<std::vector<double> >& HARKIOSrcToGNUPlotVector(harkio::M2PG::SRC_LIST_TYPE* src);
};



} //namespace
#endif

