#include <fstream>
#include <complex>
#include <sys/types.h>
#include <unistd.h>
#include <cmath>
#include <cstring>
#include <QApplication>
#include <QCoreApplication>
#include <QString>
#include <libharkio/textdata.h>
#include "UITFData.h"
#include "UIGraph.h"
#include "UITFDocument.h"
#include "Matrix.h"
#include "common.h"

#ifdef GPL
	#include <fftw.h>
#elif defined(HARKLICENSE)
	#include "fft.hpp"
#else
	#define HARKLICENSE
	#include "fft.hpp"
#endif

using namespace std;
using namespace harkio;
using namespace boost;

namespace HT {
namespace GSS {

UIData::UIData(UIDocument *document) {
	HT::UIData::document = document;
	tsp_canceled=false;
}

UIData::UIData() {
	tsp_canceled=false;
}

UIData::~UIData() {
	IFDEBUG(DP("# UIData Destructor"));
}

UIDocument* UIData::getDocument() {
	return NULL; //dynamic_cast<HT::GSS::UIDocument*>(document);
}

TFMatrix<HT_COMPLEX_DATA_TYPE>* UIData::IFFT(TFMatrix<HT_COMPLEX_DATA_TYPE>& matrix, int FFT, int NMic, int NSrc) {

	// 結果を戻す配列
	TFMatrix<HT_COMPLEX_DATA_TYPE> *result_matrix = new TFMatrix<HT_COMPLEX_DATA_TYPE>(DM3, FFT, NMic, NSrc);

#ifdef GPL
// 逆FFT
	fftw_complex *in, *out;
	fftw_plan p;

	TFMatrix<HT_COMPLEX_DATA_TYPE> copy_matrix(DM3, FFT, NMic, NSrc); // 周波数ラインを元に戻す配列
	// matrix -> result へ(半分 + 1)コピー
	copy_matrix.range(DM3, FRQ, 0, FFT/2) = matrix.range(DM3, FRQ, 0, FFT/2);

	// 残り半分へ、複素共役をとってコピー
	for (int i=1; i < FFT/2; i++) { // 1から開始
//		matrix.range(DM3, FRQ, i, i).conj();
		copy_matrix.range(DM3, FRQ, FFT-i, FFT-i) = matrix.range(DM3, FRQ, i, i);
		copy_matrix.range(DM3, FRQ, FFT-i, FFT-i).conj();
	}

	in  = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * FFT);
	out = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * FFT);
	p   = fftw_create_plan(FFT, FFTW_BACKWARD, FFTW_ESTIMATE); 	// 逆FFT

	for(int mic =0; mic < NMic; mic++) {
		for(int src =0; src < NSrc; src++) {
			// fftw 領域へコピー
			for(int fft =0; fft < FFT; fft++) {
				HT_COMPLEX_DATA_TYPE *mat = &(copy_matrix.matrix[fft][mic][src]);
				in[fft].re = mat->real();
				in[fft].im = mat->imag();
			}
			// Execute
			fftw_one(p, in, out);
			// 結果の取得
			for(int fft =0; fft < FFT; fft++) { // 同じループ
				result_matrix->matrix[fft][mic][src] = HT_COMPLEX_DATA_TYPE(out[fft].re, out[fft].im);
			}
		}
	}
	fftw_destroy_plan(p);
	fftw_free(in);
	fftw_free(out);
#endif
#ifdef HARKLICENSE
	HRI::rfft<HT_DATA_TYPE> fft_obj(FFT);
	for (int mic = 0; mic < NMic; mic++) {
		for (int src = 0; src < NSrc; src++) {
			for (int fft = 0; fft < FFT/2.0+1; fft++) {
				fft_obj.A[fft] = matrix.matrix[fft][mic][src];
			}
//			fft_obj.analyze();
			fft_obj.synthesize();
			for (int fft = 0; fft < FFT; fft++) {
				result_matrix->matrix[fft][mic][src] = HT_COMPLEX_DATA_TYPE(fft_obj.a[fft], 0);
			}
		}
	}
#endif
	return result_matrix;
}

/*
 * TFファイルの計算して作成する
 */
int UIData::createTFData(const CreateTFParam& param) {

	IFDEBUG(DP("START CREATE"));

	// パラメータのコピー
	int FFT = param.NFFT;
	int Fs = param.NFS;
	int NIR = param.IMPULSE_RESPONSE_LEN;
	int NSrc;
	int NMic;

	// 新規作成の時、バージョンとファイルタイプは固定
	setVersion(V_ONE);
	setFileType(TF_CONJ);

	vector<harkio::UIMic<> > mic_list;
	vector<harkio::UISrc<> > src_list;
	TFMatrix<HT_COMPLEX_DATA_TYPE> matrix_H;
	QString arg_channel_list;

	try {
		enum FILEMODE {
			UNDEFINE, TSP, IMPULSE
		};
		FILEMODE file_mode = UNDEFINE; // 何も指定がなければ(古い仕様の場合)、インパルス応答とみなす。

		harkio::TextData files;
		harkio::TextBase *file;

		// マイクファイル読み込み
		file = files.readFile(param.mic_file);
		mic_list = ((harkio::MICARYLocationFile*) file)->mic_lists;

		// ソースファイル読み込み
		int ret;
		ret = files.checkFileIdentifier(param.src_file);

		switch (ret) {
		case harkio::TextData::TYPE::TSPListFile:
			file_mode = TSP;
			file = files.readFile(param.src_file);
			src_list = ((harkio::TSPListFile*) file)->src_list;
			((harkio::TSPListFile*) file)->atLeastVersion(harkio::TSPListFile::V1_0);
			break;
		case harkio::TextData::TYPE::ImpulseResponseListFile:
			file_mode = IMPULSE;
			file = files.readFile(param.src_file);
			src_list = ((harkio::ImpulseResponseListFile*) file)->src_list;
			((harkio::ImpulseResponseListFile*) file)->atLeastVersion(harkio::ImpulseResponseListFile::V1_0);
			break;
		default:
			THROW_CRITICAL("Unknown format file for Srcinf's file");
			break;
		}
		std::auto_ptr<harkio::TextBase> tempptr(file);
		NMic = mic_list.size();
		NSrc = src_list.size();

		IFDEBUG(DP("PARSING FLT FILE"));

		// TFMatrix生成初期化
		TFMatrix<HT_COMPLEX_DATA_TYPE> matrix(3, FFT, NMic, NSrc);
		std::string temp_dir;

		try {
			if (file_mode == TSP) {
				// 作業用テンポラリディレクトリの作成
				char temp[PATH_MAX];
				snprintf(temp, sizeof(temp) - 1, "%s/%d-XXXXXX", harktoolDir().c_str(), getpid());
				temp_dir = mkdtemp(temp);

				// エラーチェック、channel_listの作成
				vector<int> channel_list  = ((harkio::TSPListFile*)file)->channel_list;
				if (channel_list.size() && mic_list.size() != channel_list.size()) {
					QString msg = QString("TSP-FileList's Num of ChannelList:%1 is different from MICARY-LoacationFile: %2").arg(channel_list.size()).arg(mic_list.size());
					THROW_WARNING(msg);
				}
				for (uint i = 0; i < channel_list.size(); i++) {
					if (i) {
						arg_channel_list += ",";
					}
					arg_channel_list += QString("%1").arg(channel_list[i]);
				}
			}

			for (int src = 0; src < NSrc; src++) {
				string implement_file = src_list[src].implement_file;

				int find_pos = -1;
				if (file_mode != TSP) {
					// *にチャンネル番号を埋め込む
					find_pos = implement_file.find("*", 0);
					if (find_pos < 0) {
						THROW_WARNING(tr("Can't find '*' character to replace the channel No. :%1").arg(implement_file.c_str()));
					}
				}
				string impulse_base_for_tsp;
				for (int mic = 0; mic < NMic; mic++) {
					string impulseresponse = implement_file;
					if (file_mode == TSP) {

						// 途中経過をイベントループに通知
						emit createdTSP(src, NSrc, this);
						QApplication::processEvents();
						if (tsp_canceled) {
							THROW_WARNING("TSP generation canceled.");
						}

						// マイク番号が変化する毎に、チャンネル数分のFLTファイルを作成する
						if (mic == 0) {
							impulse_base_for_tsp = fullPathToFile(src_list[src].implement_file);
							QString app_dir = QCoreApplication::applicationDirPath();
							string utils_dir = app_dir.toStdString() + "/" + HARK_TOOLS_DIR;
							string original_impulse_arg;
							TSPListFile *tspfile = ((TSPListFile*) file);
							if (checkSoundFile(tspfile->originalImpulse())) {
								original_impulse_arg = " --tsporigwav ";
							} else {
								original_impulse_arg = " --tsporigflt ";
							}
							string command = utils_dir + "/" + TSP_TO_IMPULSE /* 変換コマンド */
							+ original_impulse_arg + tspfile->originalImpulse() /* オリジナルインパルス */
							+ " --sigmax " + lexical_cast<string> (tspfile->signalMax()) /* シグナルマックス */
							+ " --syncaddfreq " + lexical_cast<string> (tspfile->syncAvg()) /* 同期加算回数 */
							+ " --syncaddoffset " + lexical_cast<string> (tspfile->tspOffset()) /* TSP オフセット */
							+ " " + implement_file /* 入力音声ファイル */
							+ " " + temp_dir + "/" + impulse_base_for_tsp; /* 出力インパルス応答ファイル */
							/* チャンネルリスト */
							if (arg_channel_list != "") {
								command +=  " --outchannellist  " + arg_channel_list.toStdString();
							}

							cout << command << endl;

							int ret = system(command.c_str());
							if (WEXITSTATUS(ret)) {
								QString msg = QString("Command Error:\n%1").arg(command.c_str());
								THROW_CRITICAL(msg);
							}
						}
						impulseresponse = (boost::format("%s/%s_ch%d.flt") %temp_dir %impulse_base_for_tsp  %(mic + 1)).str();

					} else if (file_mode == IMPULSE) {
						impulseresponse.replace(find_pos, 1, boost::lexical_cast<string>(mic + 1));
					}
					// FLT Open
					ifstream ifs_flt(impulseresponse.c_str(), ios::binary);
					if (ifs_flt.fail()) {
						QString msg = QString("Can't open flt file:%1").arg(impulseresponse.c_str());
						THROW_CRITICAL(msg);
					}
					// cutstart 分ＦＬＴをシーク
					ifs_flt.seekg(((src_list[src].cut_start) - 1) * sizeof(HT_FLT_DATA_TYPE), std::ios::beg);

					// TFMatrixに fltファイルの中身を格納
					HT_FLT_DATA_TYPE buf;
					int fft = 0;
					while (!ifs_flt.eof()) {
						ifs_flt.read(reinterpret_cast<char*> (&buf),
								sizeof(HT_FLT_DATA_TYPE));
						matrix.matrix[fft][mic][src] = buf;
						fft++;
						if (fft >= FFT) {
							break;
						}
					}
				}
			}
			if (file_mode == TSP) {
				// 終了通知
				emit createdTSP(-1, NSrc, this);
				// テンポラリ削除
				std::string temp = (boost::format("rm -rf %s") % temp_dir).str();
				int ret = system(temp.c_str());
				(void)ret;
			}
		} catch (...) {
			// 終了通知
			emit createdTSP(-1, NSrc, this);
			throw;
		}

		for (int src = 0; src < NSrc; src++) {
			TFMatrix<HT_COMPLEX_DATA_TYPE> tmp_matrix;
			HT_COMPLEX_DATA_TYPE size;
			tmp_matrix = matrix;
			tmp_matrix.range(3, 2, src, src).pow(2);
			size = tmp_matrix.range(3, 2, src, src).sum();
			matrix.range(3, 2, src, src) /= size;
		}
		TFMatrix<HT_COMPLEX_DATA_TYPE> matrix_H(3, FFT, NMic, NSrc);

		IFDEBUG(DP("EXEC FFT"));
#ifdef GPL
		// FFT
		fftw_complex *in, *out;
		fftw_plan p;

		in = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * FFT);
		out = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * FFT);
		p = fftw_create_plan(FFT, FFTW_FORWARD, FFTW_ESTIMATE);

		for(int mic =0; mic < NMic; mic++) {
			for(int src =0; src < NSrc; src++) {

				// fftw 領域へコピー
				for(int fft =0; fft < FFT; fft++) {
					HT_COMPLEX_DATA_TYPE *mat = &(matrix.matrix[fft][mic][src]); in[fft].re = mat->real();
					in[fft].im = 0;
				}
				// Execute
				fftw_one(p, in, out);
				// 結果の取得
				for(int fft =0; fft < FFT; fft++) { // 同じループ
					matrix_H.matrix[fft][mic][src] = HT_COMPLEX_DATA_TYPE(out[fft].re, out[fft].im);
				}
			}
		}
		matrix_H.print();

		fftw_destroy_plan(p);
		fftw_free(in);
		fftw_free(out);
#endif
		//#ifdef HARKLICENSE
		HRI::fft<HT_DATA_TYPE> fft_obj(FFT);
		for (int mic = 0; mic < NMic; mic++) {
			for (int src = 0; src < NSrc; src++) {
				for (int fft = 0; fft < FFT; fft++) {
					fft_obj.a[fft] = matrix.matrix[fft][mic][src]; // fftwとの違いに注意
				}
				fft_obj.analyze(); // fft実行
				for (int fft = 0; fft < FFT; fft++) {
					matrix_H.matrix[fft][mic][src] = fft_obj.A[fft];
				}
			}
		}
		//	matrix_H.print();
		//#endif

		//IFDEBUG(DP("AFTER FFT"));


		// matrix_H を半分にする
		matrix_H.trim(3, FRQ, 0, FFT / 2);
		matrix_H.range(3, FRQ, FFT / 2, FFT / 2) = 0;

#if 0 //(sn) Two line added for debug
		cout << "(sn) NFreq=" << NFreq << " NMic=" << NMic << " NSrc=" << NSrc << endl;
		cout << "(sn) FFT=" << FFT << " FRQ=" << FRQ << " SRC=" << SRC << endl;
#endif

		// ----------------------------------------
		// 2012-02-28 (s.noguchi) 正規化の方法による分岐
		if (getNormalization() == CH) {
			TFMatrix<HT_COMPLEX_DATA_TYPE> matrix_temp(3, FFT / 2 + 1, NMic, NSrc);
			matrix_temp = matrix_H;
			matrix_temp.abs();			// 絶対値 複素数でもＯＫ
			matrix_temp.pow(2);			// べき乗(単項演算子)

#if 1		// 2012-0306 (s.noguchi) マイクの各要素のSQRをしてから加算する
			cout << "(NEW)チャネルで正規化" << endl;

			for (int fft = 0; fft < FFT / 2 + 1; fft++) {
				for (int src = 0; src < NSrc; src++) {
					HT_DATA_TYPE sum = 0.0;
					HT_DATA_TYPE sum_sqr = 0.0;
					for (int mic = 0; mic < NMic; mic++) {
						sum = matrix_temp.range(9, FRQ, fft, fft, MIC, mic, mic, SRC, src, src).sum().real();
						sum_sqr += sqrt(sum);
						//cout << "sum=" << sum << " sum_sqr=" << sum_sqr << endl;
					}

					// (sn)NFreqを使用したかったが、期待した値でなかったので使用不可
					HT_DATA_TYPE avg = sum_sqr / NMic;
					//cout << "avg=" << avg << " sum_sqr=" << sum_sqr << endl;

					// 上で大きの２乗をとっているのでreal()で問題ない
					if (isfinite(sqrt(sum)) && sqrt(sum)) { // 0割り防止
						matrix_H.range(6, FRQ, fft, fft, SRC, src, src) /= avg;
					}
				}
			}
#else
	 		//2012-02-22 (s.noguchi) Comment out
			cout << "(OLD)ベクトル" << endl;

			// 新しい書き方
			TFMatrix<HT_COMPLEX_DATA_TYPE> matrix_temp(3, FFT / 2 + 1, NMic, NSrc);
			matrix_temp = matrix_H;
			matrix_temp.abs();
			matrix_temp.pow(2);
			for (int fft = 0; fft < FFT / 2 + 1; fft++) {
				for (int src = 0; src < NSrc; src++) {
					HT_DATA_TYPE sum = matrix_temp.range(6, FRQ, fft, fft, SRC, src, src).sum().real();
					// 上で大きの２乗をとっているのでreal()で問題ない
					if (isfinite(sqrt(sum)) && sqrt(sum)) { // 0割り防止
						matrix_H.range(6, FRQ, fft, fft, SRC, src, src) /= sqrt(sum);
					}
				}
			}
#endif
		}
		else if (getNormalization() == CH_FRQ) {
			// 2012-02-28 (s.noguchi) チャネル+周波数で正規化
			cout << "(NEW)チャネル+周波数で正規化" << endl;

			TFMatrix<HT_COMPLEX_DATA_TYPE> matrix_temp(3, FFT / 2 + 1, NMic, NSrc);
			matrix_temp = matrix_H;
			matrix_temp.abs();			// 絶対値 複素数でもＯＫ
			matrix_temp.pow(2);			// べき乗(単項演算子)

			// 2012-03-06 (s.noguchi) マイクと周波数の各要素のSQRをしてから加算する
			for (int src = 0; src < NSrc; src++) {
				HT_DATA_TYPE sum = 0.0;
				HT_DATA_TYPE sum_sqr = 0.0;
				for (int fft = 0; fft < FFT / 2 + 1; fft++) {
					for (int mic = 0; mic < NMic; mic++) {
						sum = matrix_temp.range(9, FRQ, fft, fft, MIC, mic, mic, SRC, src, src).sum().real();
						sum_sqr += sqrt(sum);
						//cout << "sum=" << sum << " sum_sqr=" << sum_sqr << endl;
					}
				}

				// (sn)NFreqを使用したかったが、期待した値でなかったので使用不可
				HT_DATA_TYPE avg = sum_sqr / (NMic * (FFT / 2 + 1));
				//cout << "avg=" << avg << " sum_sqr=" << sum_sqr << endl;

				for (int mic = 0; mic < NMic; mic++) {
					matrix_H.range(6, MIC, mic, mic, SRC, src, src) /= avg;
				}
			}
		}
		else if (getNormalization() == CH_FRQ_DIR) {
#if 1
			// 2012-02-22 (s.noguchi) チャネル+周波数+方向で正規化
			cout << "(NEW)チャネル+周波数+方向で正規化" << endl;

			TFMatrix<HT_COMPLEX_DATA_TYPE> matrix_temp(3, FFT / 2 + 1, NMic, NSrc);
			matrix_temp = matrix_H;
			matrix_temp.abs();
			matrix_temp.pow(2);

			// 2012-03-06 (s.noguchi) マイクと周波数と方向の各要素のSQRをしてから加算する
			HT_DATA_TYPE sum = 0.0;
			HT_DATA_TYPE sum_sqr = 0.0;

			for (int src = 0; src < NSrc; src++) {
				for (int fft = 0; fft < FFT / 2 + 1; fft++) {
					for (int mic = 0; mic < NMic; mic++) {
						sum = matrix_temp.range(9, FRQ, fft, fft, MIC, mic, mic, SRC, src, src).sum().real();
						sum_sqr += sqrt(sum);
						//(sn) cout << "sum=" << sum << " sum_sqr=" << sum_sqr << endl;
					}
				}
			}

			// (sn)NFreqを使用したかったが、期待した値でなかったので使用不可
			HT_DATA_TYPE avg = sum_sqr / ((FFT / 2 + 1) * NMic * NSrc);
			//(sn) cout << "avg=" << avg << " sum_sqr=" << sum_sqr << endl;

			//この関数を使いたかったが、segvで落ちるので使用不可
			//matrix_H.range(3, FFT / 2 + 1, NMic, NSrc) /= avg;

			for (int src = 0; src < NSrc; src++) {
				for (int fft = 0; fft < FFT / 2 + 1; fft++) {
					for (int mic = 0; mic < NMic; mic++) {
						matrix_H.range(9, FRQ, fft, fft, MIC, mic, mic, SRC, src, src) /= avg;
					}
				}
			}
#else
			cout << "(OLD)マトリックスで正規化" << endl;

			// 古い書き方 (matrixの全要素の大きさで割る)
			TFMatrix<HT_COMPLEX_DATA_TYPE> matrix_temp(3, FFT / 2 + 1, NMic, NSrc);
			matrix_temp = matrix_H;
			matrix_temp.abs();
			matrix_temp.pow(2);

			HT_COMPLEX_DATA_TYPE mean = matrix_temp.mean();
			mean = sqrt(mean.real());

 #if 0 // 2012-02-21 (s.noguchi) バグ修正、コメントアウト
			matrix /= mean;
			cout << "matrix /= mean" << endl;
 #else
			matrix_H /= mean;
			cout << "(BugFix)matrix_H /= mean" << endl;
 #endif
#endif
		}
		else {
			// 2012-02-21 (s.noguchi) 正規化は何もしない
			cout << "(NEW)何もしない" << endl;
		}

		// TFMatrix
		matrix_H.conj();
		// TFMatrix完成

		// 新規作成時に、データをhakrioのコンテナにすべて移動する
		moveToHARKIOParam param;
		param.NMic = NMic;
		param.NFFT = FFT;
		param.NIR = NIR;
		param.NFS = Fs;

		vector<HT::UIMic<HT_DATA_TYPE> > mic_list2;
		vector<HT::UISrc<HT_DATA_TYPE> > src_list2;
		src_list2.resize(src_list.size());
		mic_list2.resize(mic_list.size());
		for (uint i=0; i < src_list.size(); i++) {
			src_list2[i].convertFromHarkIO(src_list[i]);
		}
		for (uint i=0; i < mic_list.size(); i++) {
			mic_list2[i].convertFromHarkIO(mic_list[i]);
		}
		param.src_list = &src_list2;
		param.mic_list = &mic_list2;
		param.tf_matrix = &matrix_H;

		moveToHARKIO(param);
	} catch (...) {
		throw;
	}
	return 0;
}

harkio::HARK* UIData::moveToHARKIO(const moveToHARKIOParam& param) {
IFDEBUG(DP("# moveToHarkIO start"));
	vector<HT::UISrc<HT_DATA_TYPE> >& src_list = *param.src_list;
	vector<HT::UIMic<HT_DATA_TYPE> >& mic_list = *param.mic_list;
	TFMatrix<HT_COMPLEX_DATA_TYPE>& tf_matrix = *param.tf_matrix;

	int NMic = param.NMic;
	int NFFT = param.NFFT;
	int NIR  = param.NIR;
	int NFS  = param.NFS;

	// ライブラリのデータコンテナに、データをコピーする
	try {
		// Data, Chunkのポインタを用意しておく
		HARK* hark = new HARK;
		setHARKIO(hark);

		Data *hgtf;

		// HGTF Data 生成
		hgtf = new HGTF();
		hark->addData(hgtf);

		// FMT Chunk　生成 (SM,GTFの区別)
		FMT *fmt;
		if (getFileType() == GSS::UIData::SM) {
			fmt  = new FMT(FMT::V1_0, FMT::SM, NMic, NFFT, NIR, NFS);
		} else {
			fmt  = new FMT(FMT::V1_0, FMT::GTF, NMic, NFFT, NIR, NFS);
		}
		// IR インパルスレスポンスは、getNFFT()と同じ
		hgtf->addChunk(fmt);
		FMT::MIC_LIST_TYPE *mic = fmt->newMic();
		for(unsigned i=0; i < mic_list.size(); i++) { // mic_listからコピー
			(*mic)[i].x = mic_list[i].x;
			(*mic)[i].y = mic_list[i].y;
			(*mic)[i].z = mic_list[i].z;
		}
		// 以下、ＳＭとＴＦで処理を分けているが、今のところ内部構造に違いは無い
		if (getFileType() == GSS::UIData::SM) {
			// SM Chunk 生成
			harkio::SM* sm = new harkio::SM(harkio::SM::V1_0, fmt, src_list.size(), NFFT / 2 + 1);
			hgtf->addChunk(sm);

			// SM Chunk に音源位置座標コンテナを生成して、src_list()からデータをセット
			harkio::SM::SRC_LIST_TYPE *src = sm->getSrc();
			for (unsigned int srcn=0; srcn < src_list.size(); srcn++) {
				(*src)[srcn].x = src_list[srcn].x;
				(*src)[srcn].y = src_list[srcn].y;
				(*src)[srcn].z = src_list[srcn].z;
			}

			// SM Chunk に伝達関数コンテナを生成して、getMatrix()からデータをセット
			harkio::SM::SM_LIST_TYPE *tf = sm->getSM();
			for(int fft=0; fft < NFFT / 2 + 1; fft++) {
				for(uint src=0; src < src_list.size(); src++) {
					for(uint mic=0; mic < mic_list.size(); mic++) {
						HT_COMPLEX_DATA_TYPE* value = &tf_matrix.matrix[fft][mic][src];
						(*tf)[src][mic][fft].real() = value->real();
						(*tf)[src][mic][fft].imag() = value->imag();
					}
				}
			}
			// SM Chunkに周波数インデックスへ、
			vector<int> *fi = sm->getFI();
			for(int fft=0; fft < getNFreq(); fft++) {
				(*fi)[fft] = fft;
			}
		} else {
			// GTF Chunk 生成
			GTF *gtf = new GTF(GTF::V1_0, fmt, src_list.size(), NFFT / 2 + 1);
			hgtf->addChunk(gtf);

			// GTF Chunk に音源位置座標コンテナを生成して、src_list()からデータをセット
			GTF::SRC_LIST_TYPE *src = gtf->getSrc();
			for (uint srcn=0; srcn < src_list.size(); srcn++) {
				(*src)[srcn].x = src_list[srcn].x;
				(*src)[srcn].y = src_list[srcn].y;
				(*src)[srcn].z = src_list[srcn].z;
			}
			// GTF Chunk に伝達関数コンテナを生成して、getMatrix()からデータをセット
			GTF::TF_LIST_TYPE *tf = gtf->getTF();
			for(int fft=0; fft < NFFT / 2 + 1; fft++) {
				for(uint src=0; src < src_list.size(); src++) {
					for(uint mic=0; mic < mic_list.size(); mic++) {
						HT_COMPLEX_DATA_TYPE *value = &tf_matrix.matrix[fft][mic][src];
						(*tf)[src][mic][fft].real() = value->real();
						(*tf)[src][mic][fft].imag() = value->imag();
					}
				}
			}
			// GTF Chunkに周波数インデックスへ、
			vector<int> *fi = gtf->getFI();
			for(int fft=0; fft < NFFT / 2 + 1; fft++) {
				(*fi)[fft] = fft;
			}
		}
		// SPEC Chunkの設定(ここでは内容は設定しない)
		SPEC *spec = new SPEC(SPEC::V1_0);
		hgtf->addChunk(spec);
		return hark;
	} catch(std::exception& expt) {
		setHARKIO(NULL);
		harkio::Error* err = dynamic_cast<harkio::Error*>(&expt);
		if (err) {
			THROW_CRITICAL(err->getMsg().c_str());
		}
		throw;
	}
}

int UIData::saveTFFile(std::string fullpath_file) {
IFDEBUG(DP("# saveTFFile start"));
	try {
		// ファイルへ保存
		harkio::HARK *hark = this->getHARKIO();
		if(!hark) {
			THROW_CRITICAL("Data is not harkio's data");
		}
		hark->writeFile(fullpath_file.c_str());

	} catch(harkio::Error &e) {
		THROW_CRITICAL(e.getMsg().c_str());
	}
	IFDEBUG(DP("saveTFFile end"));
	return 0;
}

} //namespace GSS
} //namespace
